/***********************************************************************************
 *                                                            Autor: Sawomir Dybiec
 * Wzorzec:         fabryka prototypw (rejestrowanie obiektw)
 * 
 * Zastosowanie: Wzorca tego uywamy, kiedy mamy informacje o typie obiektu ktry 
 * chcemy utworzy. Programista w momencie pisania kodu nie moe przewidzie, jakiej
 * klasy obiekt naley utworzy. Informacja ta znana bdzie dopiero podczas wykonywania 
 * programu(np. na podstawie informacji zawartych w pliku). 
 *  
 * Opis: Wzorzec ten jest waciwie poczeniem wzorca fabryki obiektw oraz wzorca
 * prototypu. Uwaga fabryka prototypw czsto jest rwnie singletonem. Zachcam do
 * zapoznania si z wymienionymi przeze mnie wzorcami.
 * 
 * Przykad: Zamy, e posiadamy klasy suce do przetwarzania plikw graficznych,
 * jednak inna klasa odpowiada za obsug obrazw zapisanych w formacie GIF, a innej
 * klasy uywamy do obrazw zapisanych jako JPG. W momencie tworzenia kodu nie wiemy,
 * jakiego typu bdzie plik ktry zechce przetworzy uytkownik naszej biblioteki.
 * W tym momencie z pomoc przychodzi nam wzorzec fabryki prototypw.
*/

#include <iostream>
#include <vector>
#include <assert.h>

using namespace std;// dla wygody, eby nie pisa wszdzie std

//Typ pliku bdzie reprezentowany poprzez obiekt string zawierajcy jego rozszerzenie 
typedef std::string IMG_TYPE;

//Funkcja pomocnicza zwracajca nam rozszerzenie pliku. 
string getFileExtension(const string & filename)
{
	string w;
	int a;
	a=filename.find_last_of(".");
	if (a==-1 || a==static_cast<int>(filename.size())-1)
	{
		//Bd plik nie posiada rozszerzenia
		assert(0);	
	}
	w.assign(filename, a+1, filename.size()- a-1);
	//Zwracamy rozszerzenie zapisane tylko duymi literami
	std::transform(w.begin(), w.end(), w.begin(), ::toupper);
	return w;
}

//Klasa reprezentujca dowolny obrazek.
class Image 
{
public:
   //Metoda zwracajca nam wskanik na konkretn instancje obiektu Image.
   virtual Image* Clone()=0;
   virtual string getName()=0;
   /*
    * W tym miejscu definiujemy metody obsugujce obrazek. Oczywicie s to take
    * metody wirtualne.
    */ 
};
/*
* Klasa reprezentujca obrazek zapisany w formacie JPG. Jest to jednoczenie
* implementacja wzorca prototyp.
*/
class ImageJPG : public Image {
   //Funkcja zwracajca nam wskanik na nowy obiekt klasy ImageJPG.
   Image* Clone() 
   {
      return new ImageJPG(*this); 
   }
   //Funkcja zwracajca srting z opisem klasy
   string getName()
   {
      return string("Klasa przetwarzajca pliki z rozszerzeniem JPG");
   }
};

//Podobnie jak klasa ImageJPG.
class ImageGIF : public Image {
   Image* Clone() 
   {
      return new ImageGIF(*this); 
   }
   string getName()
   {
      return string("Klasa przetwarzajca pliki z rozszerzeniem GIF");
   }
};
/*
* Klasa pomocnicza wica prototyp klasy przetwarzajcej obraz z odpowiedni wartoci
* pola IMG_TYPE.
*/
class ImagePrototype
{
public:
   ImagePrototype(IMG_TYPE t, Image *p):type(t),prototype(p){}
   IMG_TYPE type;
   Image *prototype;
};
//Klasa, ktrej uywamy do tworzenia obiektw.
class ImagePrototypeFactory
{ 
   std::vector<ImagePrototype> prototypes;
public:
   /*
   * Funkcja suca do dodawania nowych prototypw klas do przetwarzania obrazw.
   * Prototypy s powizane z odpowiedni wartoci pola IMG_TYPE.
   */
   void addPrototype( IMG_TYPE img_type, Image * prototype )
   {
      ImagePrototype ip = ImagePrototype(img_type, prototype);
      prototypes.push_back( ip );
   }
   //Funkcja zwracajca wskanik  na obiekt, ktry potrafi obsugiwa zadany typ obrazka.
   Image *getImage( IMG_TYPE img_type )
   {
      int a = static_cast<int>(prototypes.size());
      for(int i=0; i < a; ++i)
         if(prototypes[i].type == img_type)
            return prototypes[i].prototype->Clone();

      //Bd nieznany typ obrazka
      return NULL;
   }
};

int main ()
{	
   //Tworzymy fabryk, ktra bdzie konstruowa odpowiedni obiekt w zalenoci od typu pliku.
   ImagePrototypeFactory ipf;
   Image * a;
   //Dodajemy prototypy klas, potraficych przetwarza odpowiednie pliki.
   ipf.addPrototype("GIF", new ImageGIF());
   ipf.addPrototype("JPG", new ImageJPG());
	
   //Prosimy Fabryk prototypw o wskanik na obiekt klasy potraficy przetwarza plik xxx.gif
   a = ipf.getImage(getFileExtension("xxx.gif"));
   //Prosimy klas o przedstawienie si :)
   cout << a->getName() << endl;

   return 0;
  
  /*
  * Wyjcie:
  * Klasa przetwarzajca pliki z rozszerzeniem GIF
  */
}

 /*
 * Przy wasnej implementacji tego wzorca, naley pamita o mechanizmie obsugi wyjtkw,
 * ktry tutaj zosta pominity.
 * Innym czstym przykadem uycia tego wzorca jest zapis stanu naszego programu do pliku,
 * wtedy podobnie jak w wyej wymienionym przykadzie dla kadej klasy definiujemy
 * odpowiedni identyfikator. Zapis kadego obiektu do pliku poprzedzony jest zapisaniem
 * identyfikatora klasy do ktrej naley obiekt. Nastpnie przy odczycie przekazujemy ten
 * identyfikator do odpowiedniej funkcji fabryki prototypw(w pokazanym przeze mnie
 * przykadzie bya to funkcja getImage()), ktra to funkcja zwraca nam wskazanie na nowo
 * utworzony obiekt klasy powizanej z tym identyfikatorem.
 * Wan cech zaprezentowanego przeze mnie wzorca jest atwo rozbudowy programu o dalsze klasy.
 * Nowo napisan klas(musi ona implementowa wzorzec prototypu) rejestrujemy w naszej fabryce poprzez
 * wywoanie funkcji addPrototype.
 */
 